@@ -290,15 +290,65 @@ def connect_weakrefable(*n, **kw):
     kw['check_same_thread'] = False
     return sqlite3.connect(*n, **kw)
 
+class ConnPool(object):
+    def __init__(self, create_fn):
+        self._create_fn = create_fn
+        self._conns = []
+        self._cond = threading.Condition()
+        self._total_conns = 0
+        self._closed = False
+        self._stack_closed_location = None
+
+    @contextlib.contextmanager
+    def get_conn(self):
+        conn = None
+        try:
+            with self._cond:
+                if self._closed:
+                    raise Exception("Closed! (at %s)" % (self._stack_closed_location,))
+
+                if not self._conns:
+                    conn = self._create_fn()
+                    self._total_conns += 1
+                else:
+                    conn = self._conns.pop()
+            assert conn is not None
+            yield conn
+        except:
+            if conn is not None:
+                # do this first
+                with self._cond:
+                    self._total_conns -= 1
+                    self._cond.notify_all()
+                conn.close()
+            raise
+        else:
+            with self._cond:
+                self._conns.append(conn)
+                self._cond.notify_all()
+
+    def close(self):
+        with self._cond:
+            if self._closed:
+                return
+
+            self._closed = True
+            self._stack_closed_location = ''.join(traceback.format_stack())
+
+            while self._total_conns != len(self._conns):
+                self._cond.wait()
+
+        for conn in self._conns:
+            conn.close()
+
 class _Directory(IterableDirectory):
     def _get_to_iter(self, mutate, fs, path):
         refreshed = True
 
-        conn = fs._get_db_conn()
-
         path_key = str(path.normed())
 
-        with trans(conn, fs._db_lock, is_exclusive=mutate), contextlib.closing(conn.cursor()) as cursor:
+        with fs._get_db_conn() as conn, \
+             trans(conn, fs._db_lock, is_exclusive=mutate), contextlib.closing(conn.cursor()) as cursor:
             cursor.execute("SELECT name, (SELECT md FROM md_cache WHERE path_key = norm_join(md_cache_entries.path_key, md_cache_entries.name)) FROM md_cache_entries WHERE path_key = ?",
                            (path_key,))
 
@@ -334,7 +384,8 @@ class _Directory(IterableDirectory):
             if not entries_names:
                 entries_names.append(EMPTY_DIR_ENT)
 
-            with trans(conn, fs._db_lock, is_exclusive=True), contextlib.closing(conn.cursor()) as cursor:
+            with fs._get_db_conn() as conn, \
+                 trans(conn, fs._db_lock, is_exclusive=True), contextlib.closing(conn.cursor()) as cursor:
                 new_stat_num = fs._get_stat_num(cursor, path_key)
                 if stat_num == new_stat_num:
                     # Cache the names we downloaded
@@ -548,7 +599,12 @@ class StreamingFile(object):
                 pass
 
         if self.cached_file is None:
-            self.cached_file = tempfile.TemporaryFile()
+            try:
+                self.cached_file = tempfile.TemporaryFile()
+            except Exception:
+                log.warning("failed to create temporary file",
+                            exc_info=True)
+                return
             is_temp = True
         else:
             is_temp = False
@@ -572,7 +628,8 @@ class StreamingFile(object):
     def _wait_for_range(self, offset, size):
         with self.cond:
             while True:
-                assert self.cached_file is not None
+                if self.cached_file is None:
+                    return True
 
                 if self.stored >= offset + size or self.eof is not None:
                     return False
@@ -584,7 +641,8 @@ class StreamingFile(object):
 
     def _should_wait(self, offset, size):
         with self.cond:
-            assert self.cached_file is not None
+            if self.cached_file is None:
+                return False
 
             # we already have the data, so we can "wait"
             if self.stored >= offset + size or self.eof is not None:
@@ -652,15 +710,24 @@ class StreamingFile(object):
                 self.stop_signal.set()
                 if self.thread is not None:
                     self.thread.join()
-                self.cached_file.close()
-                if isinstance(getattr(self.cached_file, 'name', None), str):
+                if self.cached_file is not None:
+                    # Generally close should never throw an exception
+                    # but in the case of using TemporaryFile it can
                     try:
-                        fn = os.path.join(self.cache_folder,
-                                          '%s.bin' % (self._stat.rev))
-                        os.rename(self.cached_file.name, fn)
+                        self.cached_file.close()
+                    except FileNotFoundError:
+                        pass
                     except Exception:
-                        log.exception("Unexpected failure to unlink lock file")
-                self.cached_file = None
+                        log.warning("failed to close temporary file",
+                                    exc_info=True)
+                    if isinstance(getattr(self.cached_file, 'name', None), str):
+                        try:
+                            fn = os.path.join(self.cache_folder,
+                                              '%s.bin' % (self._stat.rev))
+                            os.rename(self.cached_file.name, fn)
+                        except Exception:
+                            log.exception("Unexpected failure to unlink lock file")
+                    self.cached_file = None
 
 class NullFile(object):
     def __init__(self, id_):
@@ -685,26 +752,38 @@ class SQLiteFrontFile(PositionIO):
         PositionIO.__init__(self)
 
         self._backfile = backfile
-        self._local = threading.local()
-        (fd, self._file_path) = tempfile.mkstemp()
-        os.close(fd)
+        self._pool = ConnPool(self._create_db_conn)
+        self._file_path = None
 
-        self._db_file = "file://%s" % (self._file_path,)
+        stat = self._backfile.stat()
+
+        try:
+            (fd, self._file_path) = tempfile.mkstemp()
+            os.close(fd)
+            db_path = "file://%s" % (self._file_path,)
+            self._setup_db(db_path, '?', stat)
+        except Exception:
+            log.warning("Error creating temporary file for writes, using memory only", exc_info=True)
+            self._clear_db()
+            self._pool = ConnPool(self._create_db_conn)
+            self._file_path = None
+            db_path = "file:dropboxsff-%d?mode=memory" % (id(self),)
+            self._setup_db(db_path, '&', stat)
+
+    def _setup_db(self, db_path, appender, stat):
+        self._db_file = db_path
 
         use_shared_cache = True
         if use_shared_cache:
-            self._db_file += "?cache=shared"
+            self._db_file += appender + "cache=shared"
             # Application locking is only necessary in shared cache mode
             # otherwise SQLite will do locking for us
             self._db_lock = SharedLock()
         else:
             self._db_lock = None
 
-        stat = self._backfile.stat()
-
-        conn = self._init_db()
-
-        with trans(conn, self._db_lock, is_exclusive=True), \
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock, is_exclusive=True), \
              contextlib.closing(conn.cursor()) as cursor:
             self._update_write_md(cursor, stat.size, stat.ctime, stat.mtime)
 
@@ -720,27 +799,25 @@ class SQLiteFrontFile(PositionIO):
                            toupdate)
 
     def replace_underlying(self, new_backfile):
-        stat = self.stat()
+        oldbackfile = self._backfile
+
         new_stat = new_backfile.stat()
 
         # basic sanity check
-        assert stat.size == new_stat.size
+        assert oldbackfile.stat().size == new_stat.size
 
-        conn = self._get_db_conn()
-
-        with trans(conn, self._db_lock, is_exclusive=True), \
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock, is_exclusive=True), \
              contextlib.closing(conn.cursor()) as cursor:
             self._update_write_md(cursor, new_stat.size, new_stat.ctime, new_stat.mtime)
-
-            oldbackfile = self._backfile
             self._backfile = new_backfile
-            oldbackfile.close()
+
+        oldbackfile.close()
 
     def stat(self):
-        conn = self._get_db_conn()
-
         stat_dict = {}
-        with trans(conn, self._db_lock), contextlib.closing(conn.cursor()) as cursor:
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock), contextlib.closing(conn.cursor()) as cursor:
             cursor.execute("SELECT name, value FROM md WHERE name IN ('size', 'mtime', 'ctime')")
             for (name, value) in cursor:
                 if name in ["mtime", "ctime"]:
@@ -753,20 +830,23 @@ class SQLiteFrontFile(PositionIO):
         return Stat(type=r.type, id=r.id, rev=None, attrs=REQUIRED_ATTRS,
                     **stat_dict)
 
+    def _clear_db(self):
+        self._pool.close()
+        if self._file_path is not None:
+            try:
+                os.unlink(self._file_path)
+            except Exception:
+                log.warning("Error unlinking dirty cache file",
+                            exc_info=True)
+
     def close(self):
         if self.closed:
             return
-        try:
-            os.unlink(self._file_path)
-        except Exception:
-            log.warning("Error unlinking dirty cache file",
-                        exc_info=True)
+        self._clear_db()
         self._backfile.close()
         super().close()
 
-    def _init_db(self):
-        conn = self._get_db_conn()
-
+    def _init_db(self, conn):
         conn.executescript("""
         CREATE TABLE IF NOT EXISTS blocks
         ( blkidx INTEGER PRIMARY KEY
@@ -783,17 +863,14 @@ class SQLiteFrontFile(PositionIO):
         return conn
 
     def _create_db_conn(self):
-        return connect_weakrefable(self._db_file, uri=True)
+        return self._init_db(connect_weakrefable(self._db_file, uri=True))
 
     def _get_db_conn(self):
-        conn = getattr(self._local, 'conn', None)
-        if conn is None:
-            conn = self._local.conn = self._create_db_conn()
-        return conn
+        return self._pool.get_conn()
 
     def is_dirty(self):
-        conn = self._get_db_conn()
-        with trans(conn, self._db_lock), contextlib.closing(conn.cursor()) as cursor:
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock), contextlib.closing(conn.cursor()) as cursor:
             cursor.execute("SELECT EXISTS(SELECT * FROM blocks)")
             if cursor.fetchone()[0]:
                 return True
@@ -859,8 +936,8 @@ class SQLiteFrontFile(PositionIO):
         return toret
 
     def pread(self, size, offset):
-        conn = self._get_db_conn()
-        with trans(conn, self._db_lock), \
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock), \
              contextlib.closing(conn.cursor()) as cursor:
             return self._pread(cursor, size, offset)
 
@@ -908,8 +985,8 @@ class SQLiteFrontFile(PositionIO):
         return len(data)
 
     def pwrite(self, data, offset):
-        conn = self._get_db_conn()
-        with trans(conn, self._db_lock, is_exclusive=True), \
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock, is_exclusive=True), \
              contextlib.closing(conn.cursor()) as cursor:
             return self._pwrite(cursor, data, offset)
 
@@ -919,8 +996,8 @@ class SQLiteFrontFile(PositionIO):
         if not blkidx_start_offset:
             blkidx_start -= 1
 
-        conn = self._get_db_conn()
-        with trans(conn, self._db_lock, is_exclusive=True), \
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock, is_exclusive=True), \
              contextlib.closing(conn.cursor()) as cursor:
             cursor.execute("SELECT value FROM md WHERE name = 'size'")
             cur_size = int(cursor.fetchone()[0])
@@ -1060,8 +1137,6 @@ class CachedFile(object):
                             to_save = tempfile.NamedTemporaryFile(dir=self._fs._cache_folder)
                             self._upload_now.seek(0)
                             shutil.copyfileobj(self._upload_now, to_save)
-                            # TODO: replace self._upload_now parent's backing file with
-                            #       StreamingFile() of local cached version
                             fn = '%s.bin' % (new_stat.rev,)
                             p = os.path.join(self._fs._cache_folder, fn)
                             # Unlink existing file since new one is definitely complete
@@ -1078,7 +1153,13 @@ class CachedFile(object):
                                         exc_info=True)
                         finally:
                             if to_save is not None:
-                                to_save.close()
+                                try:
+                                    to_save.close()
+                                except FileNotFoundError:
+                                    pass
+                                except Exception:
+                                    log.warning("failed to close temporary file",
+                                                exc_info=True)
 
                     self._base_stat = new_stat
 
@@ -1305,20 +1386,7 @@ class FileSystem(object):
 
             self._db_lock = None
 
-        self._local = threading.local()
-
-        # NB: at least one conn must be held open if this is an
-        #     in-memory DB
-        self._conn_thread_stop = threading.Event()
-        self._conn_thread_started = threading.Event()
-        def conn_thread_start():
-            conn = self._init_db()
-            self._conn_thread_started.set()
-            self._conn_thread_stop.wait()
-            conn.close()
-        self._conn_thread = threading.Thread(target=conn_thread_start)
-        self._conn_thread.start()
-        self._conn_thread_started.wait()
+        self._pool = ConnPool(self._create_db_conn)
 
         self._file_cache_lock = threading.Lock()
         self._open_files_by_id = {}
@@ -1385,14 +1453,13 @@ class FileSystem(object):
             return
         self._close_prune_thread = True
         self._prune_event.set()
-        self._conn_thread_stop.set()
-        self._conn_thread.join()
         if self._watch_stop is not None:
             self._watch_stop()
         self._refresh_thread_stop = True
         self._refresh_queue.put(None)
         self._refresh_thread.join()
         self._fs.close()
+        self._pool.close()
 
     def _check_space(self, size):
         try:
@@ -1433,7 +1500,7 @@ class FileSystem(object):
                 for name in os.listdir(self._cache_folder):
                     cache_entries.append((name, os.lstat(os.path.join(self._cache_folder, name))))
 
-                cache_size = sum(st.st_size for (_, st) in cache_entries)
+                cache_size = sum(st.st_blocks * 512 for (_, st) in cache_entries)
 
                 # sort by ascending atime, descending size
                 cache_entries.sort(key=lambda name_st_pair: -name_st_pair[1].st_size)
@@ -1446,12 +1513,17 @@ class FileSystem(object):
                 # delete oldest accessed files, largest files until P<=N
                 potential_free_space = cache_size + free_space
                 for (name, st) in cache_entries:
-                    if cache_size / potential_free_space <= N:
+                    try:
+                        if cache_size / potential_free_space <= N:
+                            break
+                    except ZeroDivisionError:
+                        # NB: not sure how this happens if len(cache_entries) != 0
+                        #     but it does happen
                         break
 
                     try:
                         os.unlink(os.path.join(self._cache_folder, name))
-                        cache_size -= st.st_size
+                        cache_size -= st.st_blocks * 512
                     except Exception:
                         log.exception("Error unlinking file: %r",
                                       os.path.join(self._cache_folder, name))
@@ -1465,9 +1537,7 @@ class FileSystem(object):
                     log.warning("Error pruning cache, sleeping...", exc_info=True)
                 self._prune_event.wait(100)
 
-    def _init_db(self):
-        conn = self._create_db_conn()
-
+    def _init_db(self, conn):
         conn.executescript("""
         CREATE TABLE IF NOT EXISTS md_cache
         ( path_key TEXT PRIMARY KEY
@@ -1496,16 +1566,14 @@ class FileSystem(object):
 
     def _create_db_conn(self):
         conn = connect_weakrefable(self._db_file, uri=True)
+        # we do this here incase all references to the database died because of some error
         conn.create_function("attr_merge", 2, wrap_show_exc(attr_merge_sql))
         conn.create_function("norm_join", 2, wrap_show_exc(self._norm_join_sql))
         register_deterministic_function(conn, "file_name_norm", 1, wrap_show_exc(self._fs.file_name_norm))
-        return conn
+        return self._init_db(conn)
 
     def _get_db_conn(self):
-        conn = getattr(self._local, 'conn', None)
-        if conn is None:
-            conn = self._local.conn = self._create_db_conn()
-        return conn
+        return self._pool.get_conn()
 
     def _submit_write(self, id_, md):
         if md is not None:
@@ -1593,8 +1661,8 @@ class FileSystem(object):
 
     def _handle_changes(self, changes):
         self._statvfs_event.set()
-        conn = self._get_db_conn()
-        with trans(conn, self._db_lock, is_exclusive=True):
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock, is_exclusive=True):
             cursor = conn.cursor()
 
             if changes == "reset":
@@ -1703,9 +1771,8 @@ class FileSystem(object):
         path_key = str(path.normed())
         parent_path_key = str(path.parent.normed())
 
-        conn = self._get_db_conn()
-
-        with trans(conn, self._db_lock, is_exclusive=mutate), contextlib.closing(conn.cursor()) as cursor:
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock, is_exclusive=mutate), contextlib.closing(conn.cursor()) as cursor:
             cursor.execute("SELECT md FROM md_cache WHERE path_key = ? limit 1",
                            (path_key,))
             row = cursor.fetchone()
@@ -1745,7 +1812,8 @@ class FileSystem(object):
             except FileNotFoundError:
                 new_stat = None
 
-            with trans(conn, self._db_lock, is_exclusive=True), contextlib.closing(conn.cursor()) as cursor:
+            with self._get_db_conn() as conn, \
+                 trans(conn, self._db_lock, is_exclusive=True), contextlib.closing(conn.cursor()) as cursor:
                 # Only update metadata cache the path entry hasn't changed
                 # and it's parent hasn't changed
                 if (parent_stat_num == self._get_stat_num(cursor, parent_path_key) and
@@ -1853,8 +1921,8 @@ class FileSystem(object):
         new_path_norm = str(newpath.normed())
 
         # Invalidate cache entries for old path tree, and new path
-        conn = self._get_db_conn()
-        with trans(conn, self._db_lock, is_exclusive=True), contextlib.closing(conn.cursor()) as cursor:
+        with self._get_db_conn() as conn, \
+             trans(conn, self._db_lock, is_exclusive=True), contextlib.closing(conn.cursor()) as cursor:
             # TODO: send renamed directories to _refresh_thread
 
             # Clear new path's, new path's parent's, and old path's parent's entries
