@@ -30,6 +30,7 @@ import time
 import ssl
 import sys
 import urllib
+import urllib.parse
 import urllib.request
 
 import email.utils as eut
@@ -186,7 +187,7 @@ class _File(PositionIO):
 
         try:
             try:
-                (md, resp) = download_connection(self._fs._access_token, self._path,
+                (md, resp) = download_connection(self._fs._get_access_token(), self._path,
                                                  start=offset,
                                                  length=None if size < 0 else size)
             except HTTPError as e:
@@ -259,7 +260,7 @@ class _ReadStream(io.RawIOBase):
                 raise ValueError("closed!")
             if self._read_conn is None:
                 try:
-                    (_, self._read_conn) = download_connection(self._fs._access_token,
+                    (_, self._read_conn) = download_connection(self._fs._get_access_token(),
                                                                self._path,
                                                                start=self._start_offset)
                 except HTTPError as e:
@@ -287,9 +288,6 @@ class _ReadStream(io.RawIOBase):
 
 ApiError = dropbox.exceptions.ApiError
 
-def convert_to_dbx_timestamp(dt):
-    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")
-
 def new_files_upload(client, f, path,
                      mode=dropbox.files.WriteMode.add,
                      autorename=False,
@@ -310,7 +308,7 @@ def new_files_upload_session_finish(client,
         strict_conflict=ci.get('strict_conflict', False),
     )
     if 'client_modified' in ci:
-        commit['client_modified'] = convert_to_dbx_timestamp(ci['client_modified'])
+        commit['client_modified'] = ci['client_modified']
     return client.files_upload_session_finish(
         buf, cursor, dropbox.files.CommitInfo(**commit),
     )
@@ -467,19 +465,72 @@ def delta_thread(dbfs):
                 log.exception("failure during longpoll")
 
 class FileSystem(object):
-    def __init__(self, access_token):
-        self._access_token = access_token
+    def __init__(self, client_kwargs):
+        self._access_token = client_kwargs.get('oauth2_access_token')
+        self._access_token_expire = None
+        self._client_kwargs = client_kwargs
         self._local = threading.local()
         self._watches = []
         self._watches_lock = threading.Lock()
         self._closed = False
 
         # share this session (i.e. connection pool) across threads
-        self._db_session = self._create_session()
+        if 'session' in self._client_kwargs:
+            raise Exception("Can't override 'session' argument")
+        self._client_kwargs['session'] = self._create_session()
 
         # kick off delta thread
         threading.Thread(target=delta_thread, args=(self,), daemon=True).start()
 
+    def _get_access_token(self):
+        if (self._access_token is not None and
+            (self._access_token_expire is None or self._access_token_expire < time.time())):
+            return self._access_token
+
+        assert self._client_kwargs['oauth2_refresh_token'] is not None
+        assert self._client_kwargs['app_key'] is not None
+
+        target_host_port = ("api.dropboxapi.com", 443)
+
+        ssl_context = ssl.create_default_context()
+        ca_bundle = os.getenv("REQUESTS_CA_BUNDLE")
+        if ca_bundle is not None:
+            ssl_context.load_verify_locations(ca_bundle)
+
+        proxy = os.getenv("HTTPS_PROXY")
+        if proxy is not None:
+            o = urllib.parse.urlparse(proxy)
+            conn = http.client.HTTPSConnection(o.hostname, o.port, timeout=HTTP_TIMEOUT, context=ssl_context)
+            conn.set_tunnel(*target_host_port)
+        else:
+            conn = http.client.HTTPSConnection(*target_host_port, timeout=HTTP_TIMEOUT, context=ssl_context)
+
+        args = {
+            "grant_type" : "refresh_token",
+            "refresh_token": self._client_kwargs['oauth2_refresh_token'],
+            'client_id': self._client_kwargs['app_key'],
+        }
+
+        path = "/oauth2/token"
+
+        headers = {
+            "Content-type": "application/x-www-form-urlencoded",
+        }
+
+        conn.request("POST", path, urllib.parse.urlencode(args), headers)
+        resp = conn.getresponse()
+
+        if resp.status != 200:
+            data = resp.read()
+            raise HTTPError(resp.status)
+
+        md = json.load(resp)
+
+        self._access_token_expire = time.time() + int(md['expires_in'])
+        self._access_token = md['access_token']
+
+        return self._access_token
+
     def _create_session(self):
         session = dropbox.create_session()
 
@@ -521,8 +572,7 @@ class FileSystem(object):
     def _clientv2(self):
         toret = getattr(self._local, '_clientv2', None)
         if toret is None:
-            self._local._clientv2 = toret = dropbox.Dropbox(self._access_token,
-                                                            session=self._db_session)
+            self._local._clientv2 = toret = dropbox.Dropbox(**self._client_kwargs)
         return toret
 
     def _get_md_inner(self, path):
